# -*- coding: utf-8 -*-
"""
In this module you find the base workflow for writing out the kkr_flexfiles and
some helper methods to do so with AiiDA
"""

from aiida.orm import Code, load_node, Bool
from aiida.engine import WorkChain, ToContext, if_
from masci_tools.io.kkr_params import kkrparams
from aiida_kkr.tools.common_workfunctions import test_and_get_codenode, get_parent_paranode, update_params_wf, get_inputs_kkr
from aiida_kkr.calculations import KkrCalculation, KkrimpCalculation, VoronoiCalculation
from aiida.engine import CalcJob
from aiida.orm import CalcJobNode
from masci_tools.io.common_functions import get_Ry2eV, get_ef_from_potfile
from aiida.orm import WorkChainNode, RemoteData, StructureData, Dict, FolderData
from aiida.common.exceptions import InputValidationError
from aiida_kkr.tools.save_output_nodes import create_out_dict_node
from aiida_kkr.tools.common_workfunctions import get_username
from aiida_kkr.workflows.dos import kkr_dos_wc
from aiida_kkr.workflows.bs import set_energy_params
import os

__copyright__ = (u'Copyright (c), 2018, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.5.6'
__contributors__ = (u'Fabian Bertoldo', u'Philipp Rüßmann')

# ToDo: add more default values to wf_parameters


class kkr_flex_wc(WorkChain):
    """
    Workchain of a kkr_flex calculation to calculate the Green function with
    KKR starting from the RemoteData node of a previous calculation (either Voronoi or KKR).

    :param options: (Dict), Workchain specifications
    :param wf_parameters: (Dict), Workflow parameters that deviate from previous KKR RemoteData
    :param remote_data: (RemoteData), mandatory; from a converged KKR calculation
    :param kkr: (Code), mandatory; KKR code running the flexfile writeout
    :param impurity_info: Dict, mandatory: node specifying information
                          of the impurities in the system

    :return workflow_info: (Dict), Information of workflow results
                            like success, last result node, list with convergence behavior
    :return GF_host_remote: (RemoteData), host GF of the system
    """

    _workflowversion = __version__
    _wf_label = 'kkr_flex_wc'
    _wf_description = 'Workflow for a KKR flex calculation starting from RemoteData node of previous converged KKR calculation'

    _options_default = {
        'queue_name': '',  # Queue name to submit jobs too
        'resources': {
            'num_machines': 1
        },  # resources to allowcate for the job
        'max_wallclock_seconds': 60 * 60,  # walltime after which the job gets killed (gets parsed to KKR)}
        'custom_scheduler_commands': '',  # some additional scheduler commands
        'withmpi': True
    }  # execute KKR with mpi or without

    _wf_default = {
        'ef_shift': 0.,  # set costum absolute E_F (in eV)
        'dos_run': False,  # activate a DOS run with the parameters given in the dos_params input
        'retrieve_kkrflex':
        True,  # retrieve the kkrflex_* files or only keep them on the computer (move to KkrimpCalculation._DIRNAME_GF_UPLOAD on the remote computer's working dir), needs to use the same computer for GF writeout as for the KKRimp calculation!
    }
    # add defaults of dos_params since they are passed onto that workflow
    _wf_default['dos_params'] = kkr_dos_wc.get_wf_defaults(silent=True)

    @classmethod
    def get_wf_defaults(self):
        """
        Print and return _wf_defaults dictionary. Can be used to easily create set of wf_parameters.
        returns _wf_defaults
        """

        print(f'Version of workflow: {self._workflowversion}')
        return self._wf_default.copy()

    @classmethod
    def define(cls, spec):
        """
        Defines the outline of the workflow
        """

        # Take input of the workflow or use defaults defined above
        super(kkr_flex_wc, cls).define(spec)

        spec.input('kkr', valid_type=Code, required=False)
        spec.input('options', valid_type=Dict, required=False)
        spec.input('wf_parameters', valid_type=Dict, required=False)
        spec.input('remote_data', valid_type=RemoteData, required=True)
        spec.input('impurity_info', valid_type=Dict, required=True)
        spec.input(
            'params_kkr_overwrite',
            valid_type=Dict,
            required=False,
            help='Set some input parameters of the KKR calculation.'
        )

        # Here the structure of the workflow is defined
        spec.outline(
            cls.start,
            if_(cls.validate_input)(
                cls.set_params_flex,
                # TODO: encapsulate this in restarting mechanism (should be a base class of workflows that start calculations)
                # i.e. use base_restart_calc workchain as parent
                cls.get_flex,  # calculate host GF and kkr-flexfiles
                cls.move_kkrflex_files,  # maybe move them on the remote folder to some other directory
            ),
            cls.return_results
        )

        # ToDo: improve error codes
        spec.exit_code(
            101, 'ERROR_INVALID_INPUT_IMP_INFO', message="ERROR: the 'impurity_info' input Dict node could not be used"
        )
        spec.exit_code(
            102,
            'ERROR_INVALID_INPUT_KKR',
            message='ERROR: the code you provided for kkr does not use the plugin kkr.kkr'
        )
        spec.exit_code(103, 'ERROR_INVALID_INPUT_REMOTE_DATA', message='ERROR: No remote_data was provided as Input')
        spec.exit_code(
            104,
            'ERROR_INVALID_CALC_PARAMETERS',
            message='ERROR: calc_parameters given are not consistent! Hint: did you give an unknown keyword?'
        )
        spec.exit_code(105, 'ERROR_CALC_PARAMETERS_INCOMPLETE', message='ERROR: calc_parameters misses keys')
        spec.exit_code(
            106,
            'ERROR_KKR_CALCULATION_FAILED',
            message='ERROR: KKR calculation to write out kkrflex files unsuccessful'
        )
        spec.exit_code(
            107, 'ERROR_NO_EF_FOUND', message='ERROR: Could not extract value for Fermi level from parent calculation'
        )

        # specify the outputs
        #spec.output('remote_folder', valid_type=RemoteData)
        spec.output('workflow_info', valid_type=Dict)
        spec.output('GF_host_remote', valid_type=RemoteData)

    def start(self):
        """
        init context and some parameters
        """

        self.report(f'INFO: started KKR flex workflow version {self._workflowversion}')

        ####### init #######
        # internal para / control para
        self.ctx.abort = False
        self.ctx.exit_code = None

        # input both wf and options parameters
        options_dict = self.inputs.options.get_dict()
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
            if wf_dict == {}:
                wf_dict = self._wf_default
                self.report('INFO: using default wf parameters')
        else:
            wf_dict = self._wf_default
            self.report('INFO: using default wf parameters')
        self.report(f'INFO: wf_parameters = {wf_dict}')

        if options_dict == {}:
            options_dict = self._options_default
            self.report('INFO: using default options parameters')

        # set values, or defaults
        # ToDo: arrange option assignment differently (look at scf.py from aiida-fleur)
        self.ctx.withmpi = options_dict.get('withmpi', self._options_default['withmpi'])
        self.ctx.resources = options_dict.get('resources', self._options_default['resources'])
        self.ctx.max_wallclock_seconds = options_dict.get(
            'max_wallclock_seconds', self._options_default['max_wallclock_seconds']
        )
        self.ctx.queue = options_dict.get('queue_name', self._options_default['queue_name'])
        self.ctx.custom_scheduler_commands = options_dict.get(
            'custom_scheduler_commands', self._options_default['custom_scheduler_commands']
        )

        self.ctx.ef_shift = wf_dict.get('ef_shift', self._wf_default['ef_shift'])
        self.ctx.dos_run = wf_dict.get('dos_run', self._wf_default['dos_run'])
        self.ctx.dos_params_dict = wf_dict.get('dos_params', self._wf_default['dos_params'])
        # fill missing key, value pairs with defaults
        for k, v in self._wf_default['dos_params'].items():
            if k not in self.ctx.dos_params_dict.keys():
                self.ctx.dos_params_dict[k] = v

        self.ctx.description_wf = self.inputs.get('description', self._wf_description)
        self.ctx.label_wf = self.inputs.get('label', self._wf_label)

        self.ctx.retrieve_kkrflex = wf_dict.get('retrieve_kkrflex', self._wf_default['retrieve_kkrflex'])

        self.report(
            'INFO: use the following parameter:\n'
            'withmpi: {}\n'
            'Resources: {}\n'
            'Walltime (s): {}\n'
            'queue name: {}\n'
            'scheduler command: {}\n'
            'description: {}\n'
            'label: {}\n'.format(
                self.ctx.withmpi, self.ctx.resources, self.ctx.max_wallclock_seconds, self.ctx.queue,
                self.ctx.custom_scheduler_commands, self.ctx.description_wf, self.ctx.label_wf
            )
        )

        # return para/vars
        self.ctx.successful = True
        self.ctx.errors = []
        self.ctx.formula = ''

    def validate_input(self):
        """
        Validate input
        """

        inputs = self.inputs
        input_ok = True

        if not 'impurity_info' in inputs:
            input_ok = False
            self.ctx.exit_code = self.exit_codes.ERROR_INVALID_INPUT_IMP_INFO  # pylint: disable=no-member

        if 'remote_data' in inputs:
            input_ok = True
        else:
            input_ok = False
            self.ctx.exit_code = self.exit_codes.ERROR_INVALID_REMOTE_DATA  # pylint: disable=no-member

        # extract correct remote folder of last calculation if input remote_folder node
        # is not from KKRCalculation but kkr_scf_wc workflow
        input_remote = self.inputs.remote_data
        # check if input_remote has single KKRCalculation parent
        parents = input_remote.get_incoming(node_class=CalcJobNode)
        nparents = len(parents.all_link_labels())
        if nparents != 1:
            # extract parent workflow and get uuid of last calc from output node
            parent_workflow = input_remote.inputs.last_RemoteData
            if not isinstance(parent_workflow, WorkChainNode):
                raise InputValidationError(
                    'Input remote_data node neither output of a KKR calculation nor of kkr_scf_wc workflow'
                )
                parent_workflow_out = parent_workflow.outputs.output_kkr_scf_wc_ParameterResults
                uuid_last_calc = parent_workflow_out.get_dict().get('last_calc_nodeinfo').get('uuid')
                last_calc = load_node(uuid_last_calc)
                if not isinstance(last_calc, KkrCalculation):
                    raise InputValidationError(
                        'Extracted last_calc node not of type KkrCalculation: check remote_data input node'
                    )
                # overwrite remote_data node with extracted remote folder
                output_remote = last_calc.outputs.remote_folder
                self.inputs.remote_data = output_remote

        if 'kkr' in inputs:
            try:
                test_and_get_codenode(inputs.kkr, 'kkr.kkr', use_exceptions=True)
            except ValueError:
                error = ('The code you provided for kkr does not '
                         'use the plugin kkr.kkr')
                self.ctx.errors.append(error)
                input_ok = False
                self.ctx.exit_code = self.exit_codes.ERROR_INVALID_INPUT_KKR  # pylint: disable=no-member

        # set self.ctx.input_params_KKR
        self.ctx.input_params_KKR = get_parent_paranode(self.inputs.remote_data)

        if input_ok:
            self.report('INFO: checking inputs successful')

        return input_ok

    def set_params_flex(self):
        """
        Take input parameter node and change to input from wf_parameter and options
        """

        self.report('INFO: setting parameters ...')

        input_links = {}
        params = self.ctx.input_params_KKR
        input_dict = params.get_dict()
        para_check = kkrparams()

        # step 1: try to fill keywords
        try:
            for key, val in input_dict.items():
                para_check.set_value(key, val, silent=True)
        except:
            return self.exit_codes.ERROR_INVALID_CALC_PARAMETERS  # pylint: disable=no-member

        # step 2: check if all mandatory keys are there
        label = ''
        descr = ''
        missing_list = para_check.get_missing_keys(use_aiida=True)
        if missing_list != []:
            kkrdefaults = kkrparams.get_KKRcalc_parameter_defaults()[0]
            kkrdefaults_updated = []
            for key_default, val_default in list(kkrdefaults.items()):
                if key_default in missing_list:
                    para_check.set_value(key_default, kkrdefaults.get(key_default), silent=True)
                    kkrdefaults_updated.append(key_default)
                    missing_list.remove(key_default)
            if len(missing_list) > 0:
                self.report(f'ERROR: calc_parameters misses keys: {missing_list}')
                return self.exit_codes.ERROR_CALC_PARAMETERS_INCOMPLETE  # pylint: disable=no-member

            else:
                self.report(f'updated KKR parameter node with default values: {kkrdefaults_updated}')
                label = 'add_defaults_'
                descr = 'added missing default keys, '

        # updatedict contains all things that will be updated with update_params_wf later on
        updatedict = {}

        runopt = para_check.get_dict().get('RUNOPT', None)
        if runopt is None:
            runopt = []

        # overwrite some parameters of the KKR calculation by hand before setting mandatory keys
        if 'params_kkr_overwrite' in self.inputs:
            for key, val in self.inputs.params_kkr_overwrite.get_dict().items():
                if key != runopt:
                    updatedict[key] = val
                else:
                    runopt = val
                self.report(f'INFO: overwriting KKR parameter: {key} with {val} from params_kkr_overwrite input node')
            input_links['params_kkr_overwrite'] = self.inputs.params_kkr_overwrite

        self.report(f'INFO: RUNOPT set to: {runopt}')

        if 'wf_parameters' in self.inputs:
            if self.ctx.dos_run or self.ctx.ef_shift != 0:
                # extract Fermi energy in Ry
                remote_data_parent = self.inputs.remote_data
                parent_calc = remote_data_parent.get_incoming(link_label_filter='remote_folder').first().node
                ef = parent_calc.outputs.output_parameters.get_dict().get('fermi_energy')
                # check if ef needs to be taken from a voronoi parent
                if ef is None:
                    objects = parent_calc.outputs.retrieved.list_object_names()
                    fname = VoronoiCalculation._OUT_POTENTIAL_voronoi
                    if fname in objects:
                        with parent_calc.outputs.retrieved.open(fname) as _f:
                            ef = get_ef_from_potfile(_f)
                if ef is None:
                    return self.exit_codes.ERROR_NO_EF_FOUND  # pylint: disable=no-member

            if self.ctx.dos_run:
                # possibly remove keys which are overwritten from DOS params
                for key in ['BZDIVIDE', 'NPT2', 'EMIN', 'EMAX', 'TEMPR']:
                    if key in updatedict:
                        updatedict.pop(key)
                # now add the dos settings to the updatedict
                for key, val in self.ctx.dos_params_dict.items():
                    updatedict[key] = val
                # set the energy contour properly (i.e. changes units from eV to Ry etc)
                updatedict = set_energy_params(updatedict, ef, kkrparams())
            elif self.ctx.ef_shift != 0:
                # get Fermi energy shift in eV
                ef_shift = self.ctx.ef_shift  #set new E_F in eV
                # calculate new Fermi energy in Ry
                ef_new = (ef + ef_shift)
                self.report(
                    'INFO: ef_old + ef_shift = ef_new: {} eV + {} eV = {} eV'.format(
                        ef * get_Ry2eV(), ef_shift, ef_new * get_Ry2eV()
                    )
                )
                updatedict['ef_set'] = ef_new

        #construct the final param node containing all of the params
        if updatedict != {}:
            updatenode = Dict(updatedict)
            updatenode.label = label + 'KKRparam_flex'
            updatenode.description = descr + 'KKR parameter node extracted from parent parameters and wf_parameter and options input node.'
            paranode_flex = update_params_wf(self.ctx.input_params_KKR, updatenode, **input_links)
        else:
            paranode_flex = self.ctx.input_params_KKR
        self.ctx.flex_kkrparams = paranode_flex
        self.ctx.flex_runopt = runopt

        set_params = {k: v for k, v in paranode_flex.get_dict().items() if v is not None}
        self.report(f'INFO: Updated params= {set_params}')

    def get_flex(self):
        """
        Submit a KKRFLEX calculation
        """

        label = 'KKRFLEX calc.'
        description = 'KKRFLEX calculation to write out host GF'
        code = self.inputs.kkr
        remote = self.inputs.remote_data
        params = self.ctx.flex_kkrparams
        imp_info = self.inputs.impurity_info
        options = {
            'max_wallclock_seconds': self.ctx.max_wallclock_seconds,
            'resources': self.ctx.resources,
            'queue_name': self.ctx.queue
        }
        if self.ctx.custom_scheduler_commands:
            options['custom_scheduler_commands'] = self.ctx.custom_scheduler_commands
        inputs = get_inputs_kkr(
            code,
            remote,
            options,
            label,
            description,
            parameters=params,
            serial=(not self.ctx.withmpi),
            imp_info=imp_info
        )
        # add retrieve_kkrflex to inputs
        inputs.retrieve_kkrflex = Bool(self.ctx.retrieve_kkrflex)

        # run the KKRFLEX calculation
        self.report('INFO: doing calculation')
        flexrun = self.submit(KkrCalculation, **inputs)

        return ToContext(flexrun=flexrun)

    def move_kkrflex_files(self):
        """
        Move the kkrflex files from the remote folder to KkrimpCalculation._DIRNAME_GF_UPLOAD
        on the remote computer's working dir. This skips retrieval to the file repository and
        reduces cluttering the database.
        """
        # skip this if we want to retrieve the data to the file repository
        if not self.ctx.retrieve_kkrflex and self.ctx.flexrun.is_finished_ok:
            # get name where GF is uploaded to
            gf_upload_path = KkrimpCalculation._DIRNAME_GF_UPLOAD

            # extract computer info
            computer = self.ctx.flexrun.computer
            # computername = computer.name

            # set upload dir (get the remote username and try 5 times if there was a connection error
            # remote_user = get_username(computer)
            workdir = computer.get_workdir()  #.format(username=remote_user)
            gf_upload_path = os.path.join(workdir, gf_upload_path)

            self.report('move kkrflex files to: ' + computer.label)

            # extract absolute filepath of retrieved dir, used as source to upload kkrflex_* files from
            abspath_remote = self.ctx.flexrun.outputs.remote_folder.get_remote_path()
            abspath_tmat = os.path.join(abspath_remote, KkrimpCalculation._KKRFLEX_TMAT)
            abspath_gmat = os.path.join(abspath_remote, KkrimpCalculation._KKRFLEX_GREEN)

            # extract uuid (used as filename for remote dir)
            uuid_retrieved = self.ctx.flexrun.outputs.retrieved.uuid
            gf_upload_path = os.path.join(gf_upload_path, uuid_retrieved)

            # open connection to computer and upload files
            with computer.get_transport() as connection:
                if not connection.isdir(gf_upload_path):
                    self.report('move kkrflex_tmat and green to: ' + gf_upload_path)
                    # create directory
                    connection.makedirs(gf_upload_path, ignore_existing=True)
                    abspath_tmat_new = os.path.join(gf_upload_path, KkrimpCalculation._KKRFLEX_TMAT)
                    abspath_gmat_new = os.path.join(gf_upload_path, KkrimpCalculation._KKRFLEX_GREEN)

                    # create a symlink to to original dir to be able to find it easily
                    connection.symlink(os.path.join(abspath_remote, 'out_kkr'), os.path.join(gf_upload_path, 'out_kkr'))
                    # copy files
                    connection.copyfile(abspath_tmat, abspath_tmat_new)
                    connection.copyfile(abspath_gmat, abspath_gmat_new)
                    # remove from remote_folder and replace by a symlink to save space on remote
                    connection.remove(abspath_tmat)
                    connection.remove(abspath_gmat)
                    connection.symlink(abspath_tmat_new, abspath_tmat)
                    connection.symlink(abspath_gmat_new, abspath_gmat)
                else:
                    self.report(
                        'dir exsists already, skip moving the kkrflex_tmat and green files to: ' + gf_upload_path
                    )

    def return_results(self):
        """
        Return the results of the KKRFLEX calculation.
        This should run through and produce output nodes even if everything failed,
        therefore it only uses results from context.
        """

        if self.ctx.exit_code is not None:
            return self.ctx.exit_code

        # capture error of unsuccessful flexrun
        if not self.ctx.flexrun.is_finished_ok:
            self.ctx.successful = False
            return self.exit_codes.ERROR_KKR_CALCULATION_FAILED  # pylint: disable=no-member

        # create dict to store results of workflow output
        outputnode_dict = {}
        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['workflow_version'] = self._workflowversion
        outputnode_dict['withmpi'] = self.ctx.withmpi
        outputnode_dict['resources'] = self.ctx.resources
        outputnode_dict['max_wallclock_seconds'] = self.ctx.max_wallclock_seconds
        outputnode_dict['queue_name'] = self.ctx.queue
        outputnode_dict['custom_scheduler_commands'] = self.ctx.custom_scheduler_commands
        outputnode_dict['successful'] = self.ctx.successful
        outputnode_dict['pk_flexcalc'] = self.ctx.flexrun.pk
        outputnode_dict['list_of_errors'] = self.ctx.errors

        # create results node with calcfuntion for data provenance
        outputnode = create_out_dict_node(
            Dict(dict=outputnode_dict), GF_host_remote=self.ctx.flexrun.outputs.remote_folder
        )
        outputnode.label = 'kkr_flex_wc_results'
        outputnode.description = ''

        # return Dict node containing information about previous calculation
        self.out('workflow_info', outputnode)
        # return retrieved data from kkrflex calculation
        self.out('GF_host_remote', self.ctx.flexrun.outputs.remote_folder)

        self.report('INFO: created GF writeout result nodes')

        self.report('INFO: done with KKRFLEX GF writeout workflow!\n')


#        self.report("Successful run: {}".format(has_flexrun))



# -*- coding: utf-8 -*-
"""
Plug-in to import a KKR calculation. This is based on the PwImmigrantCalculation of the aiida-quantumespresso plugin.
"""

import os
from aiida.orm import RemoteData, Dict, StructureData
#from aiida.engine.calculation.job import _input_subfolder
#from aiida.common.utils import classproperty
#from aiida.common.folders import SandboxFolder
#from aiida.common.exceptions import InputValidationError, InvalidOperation
#from aiida.common.datastructures import calc_states
#from aiida.common.links import LinkType
from .kkr import KkrCalculation
#from masci_tools.io.kkr_params import kkrparams
#from aiida_kkr.tools.common_workfunctions import structure_from_params

__copyright__ = (u'Copyright (c), 2017, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.2'
__contributors__ = ('Philipp Rüßmann')


class KkrImporterCalculation(KkrCalculation):
    """
    Importer dummy calculation for a previous KKR run

    :param remote_workdir: Absolute path to the directory where the job was run.
        The transport of the computer you link ask input to the calculation is
        the transport that will be used to retrieve the calculation's files.
        Therefore, ``remote_workdir`` should be the absolute path to the job's
        directory on that computer.
    :type remote_workdir: str
    :param input_file_names: The file names of the job's input file.
    :type input_file_name: dict with str entries
    :param output_file_name: The file names of the job's output file (i.e. the
        file containing the stdout of kkr.x).
    :type output_file_name: dict with str entries
    """

    def _init_internal_params(self):
        """
        Init internal parameters at class load time
        """
        # reuse base class function
        super(KkrImporterCalculation, self)._init_internal_params()

        # calculation plugin version
        self._CALCULATION_PLUGIN_VERSION = __version__

        # parser
        self._default_parser = 'kkr.kkrimporterparser'

    # not working at the moment
    '''
    @classproperty
    def _use_methods(cls):
        """
        Add use_structure method for KKRimporter calculations.
        """
        use_dict = KkrCalculation._use_methods
        use_dict.update({
            "structure": {
                'valid_types': StructureData,
                'additional_parameter': None,
                'linkname': 'structure',
                'docstring':
                ("Use a node that specifies the input crystal structure ")
                }
            })
        return use_dict

    def create_input_nodes(self, open_transport, input_file_names=None,
                           output_file_names=None, remote_workdir=None):
        """
        Create calculation input nodes based on the job's files.

        :param open_transport: An open instance of the transport class of the calculation's computer. See the tutorial for more information.
        :type open_transport: aiida.transport.plugins.local.LocalTransport or aiida.transport.plugins.ssh.SshTransport

        This method parses the files in the job's remote working directory to
        create the input nodes that would exist if the calculation were
        submitted using AiiDa. These nodes are:
        * a ``'parameters'`` Dict node, based on the namelists and
        their variable-value pairs;
        * ...;
        and can be retrieved as a dictionary using the ``get_inputs_dict()``
        method. *These input links are cached-links; nothing is stored by this
        method (including the calculation node itself).*

        **Keyword arguments**
        .. note:: These keyword arguments can also be set when instantiating the
        class or using the ``set_`` methods (e.g. ``set_remote_workdir``).
        Offering to set them here simply offers the user an additional
        place to set their values. *Only the values that have not yet been
        set need to be specified.*

        :param input_file_names: The file name of the job's input files (inputcard, shapefun).
        :type input_file_names: dict of str values
        :param output_file_names: The file name of the job's output files (i.e.
            the files containing the stdout of KKR, the output.000.txt, output.0.txt, output.2.txt, out_timing.txt, output potential, output nonco_angle file).
        :type output_file_names: dict of str values
        :param remote_workdir: Absolute path to the directory where the job
            was run. The transport of the computer you link ask input to the
            calculation is the transport that will be used to retrieve the
            calculation's files. Therefore, ``remote_workdir`` should be the
            absolute path to the job's directory on that computer.
        :type remote_workdir: str
        :raises aiida.common.exceptions.InputValidationError: if
            ``open_transport`` is a different type of transport than the
            computer's.
        :raises aiida.common.exceptions.InvalidOperation: if
            ``open_transport`` is not open.
        :raises aiida.common.exceptions.InputValidationError: if
            ``remote_workdir``, ``input_file_names``, and/or ``output_file_names``
            are not set prior to or during the call of this method.
        :raises aiida.common.exceptions.FeatureNotAvailable: if the input file
            uses anything which is not currently implimented in aiida-kkr.
        :raises aiida.common.exceptions.ParsingError: if there are issues
            parsing the input file.
        :raises IOError: if there are issues reading the input file.
        """
        # Make sure the remote workdir and input + output file names were
        # provided either before or during the call to this method. If they
        # were just provided during this method call, store the values.
        if remote_workdir is not None:
            self.set_remote_workdir(remote_workdir)
        elif self.get_attr('remote_workdir', None) is None:
            raise InputValidationError(
                'The remote working directory has not been specified.\n'
                'Please specify it using one of the following...\n '
                '(a) pass as a keyword argument to create_input_nodes\n'
                '    [create_input_nodes(remote_workdir=your_remote_workdir)]\n'
                '(b) pass as a keyword argument when instantiating\n '
                '    [calc = KkrImporterCalculation(remote_workdir=your_remote_workdir)]\n'
                '(c) use the set_remote_workdir method\n'
                '    [calc.set_remote_workdir(your_remote_workdir)]'
            )

        # check if necessary input file names are given and set filenames from dict
        self.set_input_file_names(input_file_names)

        # set output file names instead of using defaults
        self.set_output_file_names(output_file_names)

        # Check that open_transport is the correct transport type.
        if type(open_transport) is not self.get_computer().get_transport_class():
            raise InputValidationError(
                "The transport passed as the `open_transport` parameter is "
                "not the same transport type linked to the computer. Please "
                "obtain the correct transport class using the "
                "`get_transport_class` method of the calculation's computer. "
                "See the tutorial for more information."
            )

        # Check that open_transport is actually open.
        if not open_transport._is_open:
            raise InvalidOperation(
                "The transport passed as the `open_transport` parameter is "
                "not open. Please execute the open the transport using it's "
                "`open` method, or execute the call to this method within a "
                "`with` statement context guard. See the tutorial for more "
                "information."
            )

        # Copy the input file and psuedo files to a temp folder for parsing.
        with SandboxFolder() as folder:

            # Copy the input file to the temp folder.
            remote_path = os.path.join(self._get_remote_workdir(),
                                       self._INPUT_FILE_NAME)
            open_transport.get(remote_path, folder.abspath)

            # Parse the input file.
            local_path = os.path.join(folder.abspath, self._INPUT_FILE_NAME)
            parameters = kkrparams(params_type='kkr')
            parameters.read_keywords_from_inputcard(inputcard=local_path)

            # Create Dict node based on the namelist and link as input.
            aiida_parameters = Dict(dict=parameters.get_dict())
            self.use_parameters(aiida_parameters)

            # Create a StructureData node from inputs
            success, structuredata = structure_from_params(parameters)
            if not success:
                raise InputValidationError('Something went wrong when creating the an aiida structure from your input file. Check you input')
            self.use_structure(structuredata)

        self._set_attr('input_nodes_created', True)

    def _prepare_for_retrieval(self, open_transport):
        """
        Prepare the calculation for retrieval by daemon.

        :param open_transport: An open instance of the transport class of the
            calculation's computer.
        :type open_transport: aiida.transport.plugins.local.LocalTransport or
            aiida.transport.plugins.ssh.SshTransport

        Here, we
        * manually set the files to retrieve
        * store the calculation and all it's input nodes
        * copy the input file to the calculation's raw_input_folder in the
        * store the remote_workdir as a RemoteData output node

        """

        # Manually set the files that will be copied to the repository and that
        # the parser will extract the results from. This would normally be
        # performed in self._prepare_for_submission prior to submission.

        natom = 1000 # maximal number of atom-resolved files that are retrieved
        # TODO take actual natom value (maybe extract from number of files that are there)
        self._set_attr('retrieve_list',
                       [self._INPUT_FILE_NAME, # inputcard needed for parsing
                        self._DEFAULT_OUTPUT_FILE, # out_kkr, std shell output
                        self._NONCO_ANGLES_OUT, # nonco angles files
                        self._OUTPUT_0_INIT, self._OUTPUT_000, self._OUTPUT_2, # output files in new(er) style
                        'output.0','output.1a','output.1b','output.1c','output.2', # try to import old style output
                        self._OUT_TIMING_000, # timing file
                        self._OUT_POTENTIAL, self._SHAPEFUN  # make sure to retrieve potential and shapefun as well
                        # add Jij files etc for other run options
                        ] + [self._SHELLS_DAT] + [self._Jij_ATOM%iatom for iatom in range(1,natom+1)]
                        )
        self._set_attr('retrieve_singlefile_list', [])

        # Make sure the calculation and input links are stored.
        self.store_all()

        # Store the original input file in the calculation's repository folder.
        remote_path = os.path.join(self._get_remote_workdir(),
                                   self._INPUT_FILE_NAME)
        raw_input_folder = self.folder.get_subfolder(_input_subfolder,
                                                     create=True)
        open_transport.get(remote_path, raw_input_folder.abspath)

        # Manually add the remote working directory as a RemoteData output
        # node.
        self._set_state(calc_states.SUBMITTING)
        remotedata = RemoteData(computer=self.get_computer(),
                                remote_path=self._get_remote_workdir())
        remotedata.add_link_from(self, label='remote_folder',
                                 link_type=LinkType.CREATE)
        remotedata.store()

    def prepare_for_retrieval_and_parsing(self, open_transport):
        """
        Tell the daemon that the calculation is computed and ready to be parsed.

        :param open_transport: An open instance of the transport class of the
            calculation's computer. See the tutorial for more information.
        :type open_transport: aiida.transport.plugins.local.LocalTransport
            or aiida.transport.plugins.ssh.SshTransport

        The next time the daemon updates the status of calculations, it will
        see this job is in the 'COMPUTED' state and will retrieve its output
        files and parse the results.
        If the daemon is not currently running, nothing will happen until it is
        started again.
        This method also stores the calculation and all input nodes. It also
        copies the original input file to the calculation's repository folder.

        :raises aiida.common.exceptions.InputValidationError: if
            ``open_transport`` is a different type of transport
            than the computer's.
        :raises aiida.common.exceptions.InvalidOperation: if
            ``open_transport`` is not open.
        """

        # Check that the create_input_nodes method has run successfully.
        if not self.get_attr('input_nodes_created', False):
            raise InvalidOperation(
                "You must run the create_input_nodes method before calling "
                "prepare_for_retrieval_and_parsing!"
            )

        # Check that open_transport is the correct transport type.
        if type(open_transport) is not self.get_computer().get_transport_class():
            raise InputValidationError(
                "The transport passed as the `open_transport` parameter is "
                "not the same transport type linked to the computer. Please "
                "obtain the correct transport class using the "
                "`get_transport_class` method of the calculation's computer. "
                "See the tutorial for more information."
            )

        # Check that open_transport is actually open.
        if not open_transport._is_open:
            raise InvalidOperation(
                "The transport passed as the `open_transport` parameter is "
                "not open. Please execute the open the transport using it's "
                "`open` method, or execute the call to this method within a "
                "`with` statement context guard. See the tutorial for more "
                "information."
            )

        # Prepare the calculation for retrieval
        self._prepare_for_retrieval(open_transport)

        # Manually set the state of the calculation to "COMPUTED", so that it
        # will be retrieved and parsed the next time the daemon updates the
        # status of calculations.
        self._set_state(calc_states.COMPUTED)

    def set_remote_workdir(self, remote_workdir):
        """
        Set the job's remote working directory.

        :param remote_workdir: Absolute path of the job's remote working
            directory.
        :type remote_workdir: str
        """
        # This is the functionality as self._set_remote_workir, but it bypasses
        # the need to have the calculation state set as SUBMITTING.
        self._set_attr('remote_workdir', remote_workdir)

    def set_input_file_names(self, input_file_names):
        """
        Set the file names of the job's input file (e.g. ``'inputcard'`` etc.).
        :param input_file_names: The file names of the job's input files (inputcard, potential, etc.).
        :type input_file_names: dict

        Keys of input_file_names dict should be one or more of:

        * ``'input_file'``
        * ``'potential_file'``
        * ``'shapefun_file'``

        :note: The keys 'input_file' and 'potential_file' are mandatory!
        """
        if self._INPUT_FILE_NAMES is None:
            self.logger.info('setting input_file_names: {}'.format(input_file_names))
            self._set_attr('input_file_names', input_file_names)
            self._set_filenames_in(input_file_names)

    def set_output_file_names(self, output_file_names):
        """
        Set the file names of the job's output files (e.g. ``'output.000.txt'`` etc.).

        :param output_file_names: The dictionary of file names of file containing the job's
            outputs.
        :type output_file_names: dict

        Keys of output_file_names dict should be one or more of:

        * ``'out_file'``
        * ``'out_potential_file'``
        * ``'out_nonco_angles_file'``
        * ``'output_0_file'``
        * ``'output_000_file'``
        * ``'output_2_file'``
        * ``'timing_file'``

        :note: In a filename is not given, the default value of aiida_kkr.calculations.kkr are used
        """
        if self._OUTPUT_FILE_NAMES is None:
            if output_file_names is None: #set default values if nothing is chosen
                self._OUT_POTENTIAL = self._POTENTIAL
                self._DEFAULT_OUTPUT_FILE = KkrCalculation()._DEFAULT_OUTPUT_FILE
                self._NONCO_ANGLES_OUT = KkrCalculation()._NONCO_ANGLES_OUT
                self._OUTPUT_0_INIT = KkrCalculation()._OUTPUT_0_INIT
                self._OUTPUT_000 = KkrCalculation()._OUTPUT_000
                self._OUTPUT_2 = KkrCalculation()._OUTPUT_2
                self._OUT_TIMING_000 = KkrCalculation()._OUT_TIMING_000
                output_file_names = {'out_potential_file': self._OUT_POTENTIAL,
                                     'out_file': self._DEFAULT_OUTPUT_FILE,
                                     'out_nonco_angles_file': self._NONCO_ANGLES_OUT,
                                     'output_0_file': self._OUTPUT_0_INIT,
                                     'output_000_file': self._OUT_TIMING_000,
                                     'output_2_file': self._OUTPUT_2,
                                     'timing_file': self._OUT_TIMING_000}
            else:
                self.logger.info('setting output_file_names: {}'.format(output_file_names))
                self._set_filenames_out(output_file_names)
             # finally set ourput_file_names as class attribute
            self._set_attr('output_file_names', output_file_names)

    def _set_filenames_in(self, input_file_names):
        """
        helper function to set filenames from input_file_names dict
        also check if at least inputcard and potential are given
        """
        has_input_files = False
        if input_file_names is not None:
            has_input_files = True
            if input_file_names.get('input_file') is not None:
                self._INPUT_FILE_NAME = input_file_names.get('input_file')
            else:
                self.logger.info('KKRimporter: input_file not given!')
                has_input_files = False
            if input_file_names.get('potential_file') is not None:
                self._POTENTIAL = input_file_names.get('potential_file')
            else:
                self.logger.info('KKRimporter: potential_file not given!')
                has_input_files = False
            if input_file_names.get('shapefun_file') is not None:
                self._SHAPEFUN = input_file_names.get('shapefun_file')
            else:
                self._SHAPEFUN = KkrCalculation()._SHAPEFUN
        if not has_input_files:
            raise InputValidationError(
                'The input file_names has not been specified correctly.\n'
                'Please specify it using one of the following...\n '
                '(a) pass as a keyword argument to create_input_nodes\n'
                '    [create_input_nodes(input_file_names=your_file_names_dictionary)]\n'
                '(b) pass as a keyword argument when instantiating\n '
                '    [calc = KkrImporterCalculation(input_file_names={"input_file":"inputcard", "potential_file":"potential", ...}]\n'
                '(c) use the set_input_file_names method\n'
                '    [calc.set_input_file_name(your_file_names)]'
            )

    def _set_filenames_out(self, output_file_names):
        """
        helper function that sets the output file names from the output_file_names dictionary
        """
        if output_file_names is not None and output_file_names.get('out_potential_file') is not None:
            self._OUT_POTENTIAL = output_file_names.get('out_potential_file')
        else:
            self._OUT_POTENTIAL = self._POTENTIAL
        #raise InputValidationError('test: {}'.format(self._OUT_POTENTIAL))
        if output_file_names is not None:
            if output_file_names.get('out_file') is not None:
                self._DEFAULT_OUTPUT_FILE = output_file_names.get('out_file')
            else:
                self._DEFAULT_OUTPUT_FILE = KkrCalculation()._DEFAULT_OUTPUT_FILE
            if output_file_names.get('out_nonco_angles_file') is not None:
                self._NONCO_ANGLES_OUT = output_file_names.get('out_nonco_angles_file')
            else:
                self._NONCO_ANGLES_OUT = KkrCalculation()._NONCO_ANGLES_OUT
            if output_file_names.get('output_0_file') is not None:
                self._OUTPUT_0_INIT = output_file_names.get('output_0_file')
            else:
                self._OUTPUT_0_INIT = KkrCalculation()._OUTPUT_0_INIT
            if output_file_names.get('output_000_file') is not None:
                self._OUTPUT_000 = output_file_names.get('output_000_file')
            else:
                self._OUTPUT_000 = KkrCalculation()._OUTPUT_000
            if output_file_names.get('output_2_file') is not None:
                self._OUTPUT_2 = output_file_names.get('output_2_file')
            else:
                self._OUTPUT_2 = KkrCalculation()._OUTPUT_2
            if output_file_names.get('timing_file') is not None:
                self._OUT_TIMING_000 = output_file_names.get('timing_file')
            else:
                self._OUT_TIMING_000 = KkrCalculation()._OUT_TIMING_000


    # These value are set as class attributes in the parent class,
    # BasePwInputGenerator, but they will be different for a job that wasn't
    # run using aiida, and they will likely vary from job to job. Therefore,
    # we override the parent class's attributes using properties, whose
    # setter methods store the values as db attributes, and whose getter
    # methods retrieve the stored values from the db.

    @property
    def _INPUT_FILE_NAMES(self):
        return self.get_attr('input_file_names', None)

    @_INPUT_FILE_NAMES.setter
    def _INPUT_FILE_NAMES(self, value):
        self._set_attr('input_file_names', value)

    @property
    def _OUTPUT_FILE_NAMES(self):
        return self.get_attr('output_file_names', None)

    @_OUTPUT_FILE_NAMES.setter
    def _OUTPUT_FILE_NAMES(self, value):
        self._set_attr('output_file_names', value)
    '''



# -*- coding: utf-8 -*-
"""
Kick out core states from the potential that are above a certain threshold
"""

from aiida.engine import calcfunction
from masci_tools.io.common_functions import open_general


def kick_out_corestates(potfile, potfile_out, emin):
    """
    Read potential file and kick out all core states that lie higher than emin.
    If no core state lies higher than emin then the output potential will be the same as the input potential
    :param potfile: input potential
    :param potfile_out: output potential where some core states are kicked out
    :param emin: minimal energy above which all core states are kicked out from potential
    :returns: number of lines that have been deleted
    """
    from masci_tools.io.common_functions import get_corestates_from_potential
    from numpy import where, array

    # read core states
    nstates, energies, lmoments = get_corestates_from_potential(potfile)

    # read potential file
    with open_general(potfile) as f:
        txt = f.readlines()

    # get start of each potential part
    istarts = [iline for iline in range(len(txt)) if 'POTENTIAL' in txt[iline]]
    all_lines = list(range(len(txt)))  # index array

    # change list of core states
    for ipot in range(len(nstates)):
        if nstates[ipot] > 0:
            m = where(energies[ipot] > emin)
            if len(m[0]) > 0:
                istart = istarts[ipot]
                # change number of core states in potential
                # print(txt[istart+6])
                txt[istart + 6] = '%i 1\n' % (nstates[ipot] - len(m[0]))
                # now remove energy line accordingly
                for ie_out in m[0][::-1]:
                    m_out = where(array(all_lines) == istart + 6 + ie_out + 1)[0][0]
                    e_out = all_lines.pop(m_out)

    # find number of deleted lines
    num_deleted = len(txt) - len(all_lines)

    if num_deleted > 0:
        # write output potential
        with open_general(potfile_out, u'w') as f2:
            txt_new = []
            for iline in all_lines:
                txt_new.append(str(txt[iline]))
            f2.writelines(txt_new)

    # return number of lines that were deleted
    return num_deleted


@calcfunction
def kick_out_corestates_wf(potential_sfd, emin):
    """
    Workfunction that kicks out all core states from single file data potential that are higher than emin.
    :param potential_sfd: SinglefileData type of potential
    :param emin: Energy threshold above which all core states are removed from potential (Float)
    :returns: potential without core states higher than emin (SinglefileData)
    """
    from aiida.common.folders import SandboxFolder
    from aiida.orm import SinglefileData

    with SandboxFolder() as tmpdir:
        with tmpdir.open('potential_deleted_core_states', 'w') as potfile_out:
            with potential_sfd.open(potential_sfd.filename) as potfile_in:
                num_deleted = kick_out_corestates(potfile_in, potfile_out, emin)
        # store new potential as single file data object
        if num_deleted > 0:
            with tmpdir.open('potential_deleted_core_states', 'rb') as potfile_out:
                potential_nocore_sfd = SinglefileData(file=potfile_out)

    # return potential
    if num_deleted > 0:
        return potential_nocore_sfd
    else:
        return potential_sfd.clone()



# -*- coding: utf-8 -*-
'''
Module for the command line interface of AiiDA-KKR
'''
import click
import click_completion

from aiida.cmdline.params import options, types
from .launch import cmd_launch
from .data import cmd_data
from .visualization import cmd_plot

# Activate the completion of parameter types provided by the click_completion package
# for bash: eval "$(_AIIDA_KKR_COMPLETE=source_bash aiida-kkr)"
click_completion.init()

# Instead of using entrypoints and directly injecting verdi commands into aiida-core
# we created our own separete CLI because verdi will prob change and become
# less material science specific


@click.group('aiida-kkr', context_settings={'help_option_names': ['-h', '--help']})
@options.PROFILE(type=types.ProfileParamType(load_profile=True))
def cmd_root(profile):  # pylint: disable=unused-argument
    """CLI for the `aiida-kkr` plugin."""


# To avoid circular imports all commands are not yet connected to the root
# but they have to be here because of bash completion

cmd_root.add_command(cmd_launch)
cmd_root.add_command(cmd_data)
cmd_root.add_command(cmd_plot)



# -*- coding: utf-8 -*-
"""
Helper functions to find the cluster radius
"""

import numpy as np
from aiida.orm import StructureData
from masci_tools.io.common_functions import get_alat_from_bravais


def find_cluster_radius_old(structure, nclsmin, n_max_box=50, nbins=100):
    """
    Takes structure information (cell and site positions) and computes the minimal cluster radius needed
    such that all clusters around all atoms contain more than `nclsmin` atoms.

    :note: Here we assume spherical clusters around the atoms!

    :param structure: input structure for which the clusters are analyzed
    :param nclsmin: minimal number of atoms in the screening cluster
    :param n_max_box: maximal number of supercells in 3D volume
    :param nbins: number of bins in which the cluster number is analyzed

    :returns: minimal cluster radius needed in Angstroem
    :returns: minimal cluster radius needed in units of the lattice constant
    """

    # extract values needed from structure
    cell = np.array(structure.cell)
    pos = np.array([site.position for site in structure.sites])

    # settings for supercell box
    box = int((n_max_box / len(pos))**(1 / 3.) + 0.5)
    # print('maximal number of atoms in box (time number of atoms in unit cell):', (box*2+1)**3)

    # find all positions in the supercell
    all_pos_box = np.zeros_like(pos)
    for i in range(-box, box + 1):
        for j in range(-box, box + 1):
            for k in range(-box, box + 1):
                for site in pos:
                    tmppos = site + i * cell[0] + j * cell[1] + k * cell[2]
                    all_pos_box = np.append(all_pos_box, [tmppos], axis=0)
    all_pos_box = all_pos_box[len(pos):]

    # computer number of atoms in the clusters
    # Attention: assumes spherical clusters!
    rclsmax_ang = -1
    for site in pos:
        tmpdiff = np.sort(np.sqrt(np.sum((all_pos_box - site)**2, axis=1)))[1:]
        rmax = tmpdiff.max()
        clssizes = [len(tmpdiff[tmpdiff < i * rmax]) for i in np.linspace(0, 1, nbins)]
        rclsmax_atom = (np.linspace(0, 1, nbins) * rmax)[np.where(np.array(clssizes) < nclsmin)[0].max() + 1]
        if rclsmax_atom > rclsmax_ang:
            rclsmax_ang = rclsmax_atom

    # convert also to alat units
    rclsmax_alat = rclsmax_ang / get_alat_from_bravais(cell, structure.pbc[2])

    # now the minimal cluster radius needed to get the spherical screening clusters around the atoms larger than
    # nclsmin atoms is found and can be returned
    return rclsmax_ang, rclsmax_alat


def find_cluster_radius(structure, nclsmin=15, Rclsmax=10.):
    """
    Find cluster radius to have >=nclsmin atoms in each cluster.

    :param structure: input structure
    :param nclsmin: required minimal cluster size
    :param Rclsmax: max radius used in screening for cluster size

    :return R0: cluster radius in Ang. units
    :return Ncls_all: cluster sizes, list with the lenth of the sites
    """

    # create auxiliary structure (makes sure this also works with 2D structures)
    saux = StructureData(cell=structure.cell)
    for site in structure.sites:
        kind = structure.get_kind(site.kind_name)
        saux.append_atom(position=site.position, symbols=kind.symbols, weights=kind.weights)
    ps = saux.get_pymatgen()

    # find cluster radius that has at least the minimal number of atoms
    R0 = -1.
    neighbors_all = ps.get_all_neighbors(max(Rclsmax, np.linalg.norm(structure.cell, axis=1).max()))
    for neighbors in neighbors_all:
        dist_all = np.sort([np.linalg.norm(n.coords) for n in neighbors])
        # -2 because we take the (nclsmin-1)-th position
        # in the array (position at (0,0,0) is added automatically)
        R0 = max(R0, dist_all[nclsmin - 2])

    # consistency check
    if R0 <= 0.:
        raise ValueError('Cluster radius not found properly')

    # count number of atoms in the clusters
    neighbors_all = ps.get_all_neighbors(R0)
    Ncls_all = [1 + len(n) for n in neighbors_all]

    return R0, Ncls_all



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
In this module you find the base workflow that defines the restart
caclulation functionality.
"""

from aiida_kkr.calculations import KkrimpCalculation, VoronoiCalculation, KkrCalculation
from aiida.engine import WorkChain
from aiida.orm import CalcJobNode

__copyright__ = (u'Copyright (c), 2019, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.1'
__contributors__ = u'Philipp Rüßmann'

# similar to base restart workflow of aiida-quantumespresso


class base_restart_calc(WorkChain):
    """
    Use this base workflow in workflows that directly sumbit calculations (not workflows)

    This workflow defines the following functions that can be used to encapsulate functions that submit calculations:
      * cls.prepare_new_restart_calc
      * cls.inspect_restart_calc_done

    You can set the behavior of the restart block by setting these values (defaults after the `=`:
      * _max_iter_restart_calc = 2
      * _clean_workdir_restart_calc = False

    Then you should make sure to run prepare_new_restart_calc each time you want to do a new calculation (with restarts).

    Make sure self.ctx.last_calc is being set to the calculation you want to be able to resubmit.

    If you want to cleanup all called claculations after running you should have a call to `cleanup_all_calcs_work` in your workflow.

    usage example:
        class my_workflow(base_restart_calc):

        ...

        spec.outline(
            ...
            #here the calculation is submitted:
            cls.prepare_new_restart_calc,
            while_(cls.inspect_restart_calc_done)(
                cls.my_custom_run_function),
            # now the calculation should have finished, eventually after being restarted to overcome cluster problems
            ...
            # if we want to cleanup the work directories (`my_workflow` class needs to
            # `set _clean_workdir_restart_calc` to True if the work directories are supposed to be removed!)
            cls.cleanup_all_calcs_work
        )
    """

    _workflowversion = __version__

    _max_iter_restart_calc = 2
    _clean_workdir_restart_calc = False

    def __init__(self, *args, **kwargs):
        super(base_restart_calc, self).__init__(*args, **kwargs)

    @classmethod
    def define(cls, spec):
        super(base_restart_calc, cls).define(spec)

        # define general exit codes for restart of calculations
        spec.exit_code(
            100,
            'ERROR_ITERATION_RETURNED_NO_CALCULATION',
            message='the run_calculation step did not successfully add a calculation node to the context'
        )

    def prepare_new_restart_calc(self):
        """initialize counters etc. used within inspect_restart_calc_done"""
        self.ctx.restart_iteration = 0
        self.ctx.last_calc_restart_failed = False

    def inspect_restart_calc_done(self):
        """check if the calculation should be submitted"""
        # initial go
        if self.ctx.restart_iteration == 0:
            self.ctx.restart_iteration += 1
            return True

        #  stop after _max_iter_restart_calc iterations
        if self.ctx.restart_iteration > self._max_iter_restart_calc:
            self.cleanup_calc_work()
            return False

        # extract last calculation
        try:
            last_calculation = self.ctx.last_calc
        except:
            self.report('')
            return self.exit_codes.ERROR_ITERATION_RETURNED_NO_CALCULATION  # pylint: disable=no-member

        # now inspect last calculation
        if last_calculation.is_finished_ok:
            # calculation already succesful
            self.cleanup_calc_work()
            return False
        else:
            # check if this is the second consecutive failure
            if self.ctx.last_calc_restart_failed:
                self.cleanup_calc_work()
                return False

            # here we know already that the current calculation has failed
            self.ctx.last_calc_restart_failed = True

            # TODO: handle other known errors (maybe pause calculation until some problem is fixed)
            # TODO: maybe capture out of walltime or out of memory issues here as well (could automatically increase the options)

            # if nothing else was caught, run the calculation
            self.ctx.restart_iteration += 1
            return True

    # this is probably not working yet (check again with aiida-quantumespresso plugin)
    def cleanup_all_calcs_work(self):
        # cleanup work directories of all called calculations
        if self._clean_workdir_restart_calc:
            cleaned_calcs = []

            for called_descendant in self.calc.called_descendants:
                if isinstance(called_descendant, CalcJobNode):
                    try:
                        called_descendant.outputs.remote_folder._clean()
                        cleaned_calcs.append(called_descendant.pk)
                    except (IOError, OSError, KeyError):
                        pass

            if cleaned_calcs:
                self.report(f"cleaned remote folders of calculations: {' '.join(map(str, cleaned_calcs))}")



#!/usr/bin/env python
# coding: utf-8
"""
This module contains the band structure workflow for KKR which is done by calculating the k-resolved spectral density
also known as Bloch spectral function.
"""

from aiida.orm import Code, Dict, RemoteData, StructureData, Float, Str, WorkChainNode, load_node, CalcJobNode, ArrayData, KpointsData
from aiida.engine import WorkChain, ToContext, calcfunction
from aiida.tools.data.array.kpoints import get_explicit_kpoints_path
from aiida_kkr.tools.common_workfunctions import test_and_get_codenode, get_parent_paranode, update_params_wf, get_inputs_kkr
from aiida_kkr.calculations.kkr import KkrCalculation
from aiida_kkr.calculations.voro import VoronoiCalculation
from aiida_kkr.tools.save_output_nodes import create_out_dict_node
from aiida_kkr.tools.extract_kkrhost_noco_angles import extract_noco_angles
from masci_tools.io.kkr_params import kkrparams
from masci_tools.io.common_functions import get_Ry2eV
import numpy as np

__copyright__ = (u'Copyright (c), 2020, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.1.8'
__contributors__ = (u'Rubel Mozumder', u'Philipp Rüßmann')


class kkr_bs_wc(WorkChain):
    """
    Workchain for BandStructure calculation, starting from RemoteFolderData of the previous converged KKR calculation remote folder data

    inputs:
    :param wf_parameters: (Dict), (optional); Workchain Specifications, contains nepts, tempr, emin (in eV relative to EF), emax (in eV),
                          and RCLUSTZ (can be used to increase the screening cluster radius) keys.

    :param options: (Dict), (optional); Computer Specifications, scheduler command, parallel or serial
    :param kpoints: (KpointsData),(optional); Kpoints data type from the structure,
                                   but not mendatory as it can be extracted from structure internaly from the remote data
    :param remote_data: (RemoteData)(mendaory); From the previous kkr-converged calculation.
    :param kkr: (Code)(mendaory); KKR code specifiaction
    :param label: (Str) (optional) ; label for WC but will be found in the 'result_wf' output
                                     Dict as 'BS_wf_label' key
    :param description: (Str) (optional) : description for WC but will be found in the 'result_wf' output
                                     Dict as 'BS_wf_description' key


    returns:
    :out BS_Data : (ArrayData) ; Consist of BlochSpectralFunction, k_points (list), energy_points (list), special_kpoints(dict)
    :out result_wf: (Dict); work_chain_specifications node, BS_data node, remote_folder node
    """

    _wf_version = __version__
    _wf_label = 'kkr_BandStructure_wc'
    _wf_description = """Workflow for a bandstructure calculation starting eithe from a structure with automatic voronoi'
                        calculation or a valid RemoteData of a previous calculation."""
    _wf_default = {
        'emin': -10.0,  # start of the energy range in eV, relative to the Fermi energy
        'emax': 5.0,  # end of the energy range in eV, relative to the Fermi energy
        'nepts': 96,  # number of energy points
        'RCLUSTZ': None,  # can be used to increase the cluster radius if a value is set here
        'tempr': 50.,  # smearing temperature in K
        'kmesh': None,  # k-point integration mesh, only useful for CPA calculation
    }

    _options_default = {
        'max_wallclock_seconds': 36000,
        'resources': {
            'num_machines': 1
        },
        'withmpi': True,
        'queue_name': '',
        'prepend_text': '',
        'append_text': '',
        'additional_retrieve_list': None
    }

    @classmethod
    def get_wf_defaults(self, silent=False):
        """
        Return the default values of the workflow parameters (wf_parameters input node)
        """
        if not silent:
            print(f'Version of the kkr_bs_wc workflow: {self._wf_version}')
        return self._wf_default.copy()

    @classmethod
    def define(cls, spec):
        """
        Layout of the workflow, defines the input nodes and the outline of the workchain
        """
        super(kkr_bs_wc, cls).define(spec)
        # here inputs are defined
        spec.input(
            'wf_parameters',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(dict=cls._wf_default),
            help='Parameters of the bandstructure workflow (see output of kkr_bs_wc.get_wf_default() for more details).'
        )
        spec.input(
            'options',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(dict=cls._options_default),
            help='Computer options (walltime etc.) passed onto KkrCalculation'
        )
        spec.input(
            'remote_data',
            valid_type=RemoteData,
            required=True,
            help='Parent folder of previously converged KkrCalculation'
        )
        spec.input('kkr', valid_type=Code, required=True, help='KKRhost code, needed to run the qdos KkrCalculation')
        spec.input(
            'kpoints',
            valid_type=KpointsData,
            required=False,
            help=
            'K-points data for the calculation. If not given the seekpath library is used to find the irreducable k-points of a structure.'
        )
        spec.input('label', valid_type=Str, required=False, help='label for the workflow')
        spec.input('description', valid_type=Str, required=False, help='description for the workflow')
        spec.input(
            'initial_noco_angles',
            valid_type=Dict,
            required=False,
            help="""Initial non-collinear angles for the magnetic moments. See KkrCalculation for details.
            If this is found in the input potentially extracted nonco angles from the parent calulation are overwritten!"""
        )
        # maybe overwrite some settings from the KKRhost convergence run
        spec.input(
            'params_kkr_overwrite',
            valid_type=Dict,
            required=False,
            help='Overwrite some input parameters of the parent KKR calculation.'
        )
        # expose LDAU input node
        spec.input(
            'settings_LDAU',
            valid_type=Dict,
            required=False,
            help="""
Settings for running a LDA+U calculation. The Dict node should be of the form
    settings_LDAU = Dict(dict={'iatom=0':{
        'L': 3,         # l-block which gets U correction (1: p, 2: d, 3: f-electrons)
        'U': 7.,        # U value in eV
        'J': 0.75,      # J value in eV
        'Eref_EF': 0.,  # reference energy in eV relative to the Fermi energy. This is the energy where the projector wavefunctions are calculated (should be close in energy where the states that are shifted lie (e.g. for Eu use the Fermi energy))
    }})
    Note: you can add multiple entries like the one for iatom==0 in this example. The atom index refers to the corresponding atom in the impurity cluster.
"""
        )

        # Here outputs are defined
        spec.output('results_wf', valid_type=Dict, required=True)
        spec.output('BS_Data', valid_type=ArrayData, required=True)

        # Here outlines are being specified
        spec.outline(
            # For initialiging workflow
            cls.start,
            cls.validate_input,
            cls.set_params_BS,
            cls.get_BS,
            cls.return_results
        )
        # definition of exit code in case something goes wrong in this workflow
        spec.exit_code(161, 'ERROR_NO_INPUT_REMOTE_DATA', 'No remote_data was provided as Input')
        spec.exit_code(
            162, 'ERROR_KKRCODE_NOT_CORRECT', 'The code you provided for kkr does not use the plugin kkr.kkr'
        )
        spec.exit_code(
            163, 'ERROR_CALC_PARAMETERS_INVALID',
            'calc_parameters given are not consistent! Hint: did you give an unknown keyword?'
        )
        spec.exit_code(164, 'ERROR_CALC_PARAMETERS_INCOMPLETE', 'calc_parameters not complete')
        spec.exit_code(165, 'ERROR_BS_CALC_FAILED', 'KKR Band Structure calculation failed')
        spec.exit_code(166, 'ERROR_NO_KPOINTS_EXTRACTED', 'No K-POINTS can be extracted from the structure data')
        spec.exit_code(
            167, 'ERROR_INCORRECT_KPOINTS_EXTRACTED',
            'No K-POINTS can be extracted from the primtive structure data rather conventional structure data'
        )
        spec.exit_code(
            168, 'ERROR_INVALID_REMOTE_DATA_TPYE',
            'Input remote_data node neither output of a KKR/voronoi calculation nor of kkr_scf_wc workflow'
        )

    def start(self):
        """
        set up context of the workflow
        """
        self.report(f'INFO: started KKR Band Structure workflow version {self._wf_version}')
        wf_dict = self.inputs.wf_parameters.get_dict()
        # Count energy points only once
        if 'NPT2' in wf_dict.keys():
            npt2 = wf_dict.pop('NPT2', None)
            wf_dict['nepts'] = npt2
        # add missing default values
        for key, val in self._wf_default.items():
            if ((key not in wf_dict.keys()) and (key.swapcase() not in wf_dict.keys()) and (val is not None)):

                self.report(f'INFO: Using default wf parameter {key}: {val}')
                wf_dict[key] = val

        options_dict = self.inputs.options.get_dict()
        if options_dict == {}:
            self.report('INFO: Using default wf Options')
            options_dict = self._options_default
        self.ctx.append_text = options_dict.get('append_text', self._options_default['append_text'])
        self.ctx.prepend_text = options_dict.get('prepend_text', self._options_default['prepend_text'])
        self.ctx.additional_retrieve_list = options_dict.get(
            'additional_retrieve_list', self._options_default['additional_retrieve_list']
        )
        self.ctx.withmpi = options_dict.get('withmpi', self._options_default['withmpi'])
        self.ctx.resources = options_dict.get('resources', self._options_default['resources'])
        self.ctx.max_wallclock_seconds = options_dict.get(
            'max_wallclock_seconds', self._options_default['max_wallclock_seconds']
        )
        self.ctx.queue = options_dict.get('queue_name', self._options_default['queue_name'])
        self.ctx.custom_scheduler_commands = options_dict.get('custom_scheduler_commands', '')
        self.ctx.BS_params_dict = wf_dict
        self.ctx.BS_kkrparams = None  # is set in set_params_BS
        self.ctx.BS_kpoints = None
        self.ctx.description_wf = self.inputs.get('description', self._wf_description)
        self.ctx.label_wf = self.inputs.get('label', self._wf_label)
        self.report(
            'INFO: use the following parameter:\n'
            'withmpi: {}\n'
            'Resources: {}\n'
            'Walltime (s): {}\n'
            'queue name: {}\n'
            'scheduler command: {}\n'
            'description_wf: {}\n'
            'label_wf: {}\n'
            'BS_params: {}\n'.format(
                self.ctx.withmpi, self.ctx.resources, self.ctx.max_wallclock_seconds, self.ctx.queue,
                self.ctx.custom_scheduler_commands, self.ctx.description_wf, self.ctx.label_wf, self.ctx.BS_params_dict
            )
        )

        self.ctx.successful = True
        self.ctx.errors = []

    def validate_input(self):
        """
        validate input and find out which path ( converged kkr calc or wf ) to take
        return True means run voronoi if false run kkr directly
        """
        inputs = self.inputs
        if 'remote_data' in inputs:
            input_ok = True
        else:
            input_ok = False
            return self.exit_codes.ERROR_NO_INPUT_REMOTE_DATA  # pylint: disable=no-member
        input_remote = self.inputs.remote_data
        parents = input_remote.get_incoming(node_class=CalcJobNode)
        nparents = len(parents.all_link_labels())

        if nparents != 1:
            # extract parent workflow and get uuid of last calc from output node
            parent_workflow = input_remote.inputs.last_RemoteData
            if not isinstance(parent_workflow, WorkChainNode):
                return self.exit_codes.ERROR_INVALID_REMOTE_DATA_TPYE  # pylint: disable=no-member

            parent_workflow_out = parent_workflow.outputs.output_kkr_scf_wc_ParameterResults
            uuid_last_calc = parent_workflow_out.get_dict().get('last_calc_nodeinfo').get('uuid')
            last_calc = load_node(uuid_last_calc)

            if not isinstance(last_calc, KkrCalculation) and not isinstance(last_calc, VoronoiCalculation):
                return self.exit_code.ERROR_INVALID_REMOTE_DATA_TPYE

            # overwrite remote_data node with extracted remote folder
            output_remote = last_calc.outputs.remote_folder

            self.inputs.remote_data = output_remote

        # extract structure
        struc_kkr, _ = VoronoiCalculation.find_parent_structure(self.inputs.remote_data)
        # save if structure is an alloy
        self.ctx.struc_is_alloy = struc_kkr.is_alloy

        # To validate for kpoints
        if 'kpoints' in inputs:
            self.ctx.BS_kpoints = inputs.kpoints
            input_ok = True
            self.ctx.structure_data = 'None (kpoints taken from input)'
        else:
            #create an auxiliary structure with unique kind_names, this leads to using the input structure in the seekpath method instead of finding the primitive one
            cell = np.array(struc_kkr.cell)
            if not struc_kkr.pbc[2]:
                # 2D structure, make sure the third bravais vector points along z
                cell[2] = np.cross(cell[0], cell[1])
            saux = StructureData(cell=cell)
            for isite, site in enumerate(struc_kkr.sites):
                kind = struc_kkr.get_kind(site.kind_name)
                symbols = kind.symbols
                weights = kind.weights
                # replace old way of empty site with new 'X' kind
                if site.kind_name == 'HX' and kind.weights[0] < 1e-8:
                    symbols = 'X'
                    weights = 1.0
                saux.append_atom(
                    name='atom' + str(isite) + ':' + site.kind_name,
                    symbols=symbols,
                    weights=weights,
                    position=site.position
                )
            # use auxiliary structure inside k-point generator
            output = get_explicit_kpoints_path(saux)
            primitive_struc = output['primitive_structure']
            conventional_struc = output['conv_structure']
            kpoints_ok = True

            #check if primitive_structure and input structure are identical:
            maxdiff_cell = sum(abs(np.array(primitive_struc.cell) - np.array(saux.cell))).max()

            if maxdiff_cell > 3 * 10**-9:
                self.report(f'Error in cell : {maxdiff_cell}')
                self.report(
                    'WARNING : The structure data from the voronoi calc is not the primitive structure type and in come cases it is medatory'
                )
                self.report(f'prim: {primitive_struc.cell} {primitive_struc.sites}')
                self.report(f'conv: {conventional_struc.cell} {conventional_struc.sites}')
                self.ctx.structure_data = 'conventional_unit_cell '
            else:
                self.ctx.structure_data = 'primitive_unit_cell'

            if not kpoints_ok:
                return self.exit_codes.ERROR_INCORRECT_KPOINTS_EXTRACTED  # pylint: disable=no-member
            else:
                kpts = output['explicit_kpoints']

            self.ctx.BS_kpoints = kpts
            if isinstance(KpointsData(), type(kpts)):
                input_ok = True
            else:
                input_ok = False
                return self.exit_codes.ERROR_NO_KPOINTS_EXTRACTED  # pylint: disable=no-member

        # To validate for kkr
        if 'kkr' in inputs:
            try:
                test_and_get_codenode(inputs.kkr, 'kkr.kkr', use_exceptions=True)
            except ValueError:
                input_ok = False
                return self.exit_codes.ERROR_KKRCODE_NOT_CORRECT  # pylint: disable=no-member

        # set self.ctx.input_params_KKR
        self.ctx.input_params_KKR = get_parent_paranode(self.inputs.remote_data)
        self.report(f'The validation input_ok {input_ok}')

    def set_params_BS(self):
        """
        set kkr parameters for the bandstructure (i.e. qdos) calculation
        """
        params = self.ctx.input_params_KKR

        # maybe overwrite some inputs
        if 'params_kkr_overwrite' in self.inputs:
            self.report(f'found params_kkr_overwrite: {self.inputs.params_kkr_overwrite.get_dict()}')
            updatenode = self.inputs.params_kkr_overwrite
            updatenode.label = 'params overwrite'
            params = update_params_wf(params, updatenode)

        input_dict = params.get_dict()
        para_check = kkrparams()
        try:
            for key, val in input_dict.items():
                para_check.set_value(key, val, silent=True)
        except:
            return self.exit_codes.ERROR_CALC_PARAMETERS_INVALID  # pylint: disable=no-member
        label = ''
        descr = f'(pk - {self.inputs.remote_data.pk}, and uuid - {self.inputs.remote_data.uuid})'

        missing_list = para_check.get_missing_keys(use_aiida=True)

        if missing_list != []:
            kkrdefaults = kkrparams.get_KKRcalc_parameter_defaults()[0]
            kkrdefaults_updated = []
            for key_default, val_default in list(kkrdefaults.items()):
                if key_default in missing_list:
                    para_check.set_value(key_default, val_default)
                    kkrdefaults_updated.append(key_default)
                    missing_list.remove(key_default)
            if len(missing_list) > 0:
                self.report(f'ERROR: calc_parameters misses keys: {missing_list}')
                return self.exit_codes.ERROR_CALC_PARAMETERS_INCOMPLETE  # pylint: disable=no-member
            else:
                self.report(f'updated KKR parameter node with default values: {kkrdefaults_updated}')
                label = 'add_defaults_'
                descr = 'added missing default keys, '
        ##+++ Starts to add the NTP2, EMAX and EMIN from the
        econt_new = self.ctx.BS_params_dict
        if self.ctx.struc_is_alloy:
            if econt_new.get('kmesh', None) is None:
                econt_new['kmesh'] = [1, 1, 1]  # overwrite kmesh since the kpoints are used from the input
        kkr_calc = self.inputs.remote_data.get_incoming(node_class=KkrCalculation).first().node
        ef = kkr_calc.outputs.output_parameters.get_dict()['fermi_energy']  # unit in Ry
        self.ctx.fermi_energy = ef  ## in Ry unit

        # Set BS params
        try:
            para_check = set_energy_params(econt_new, ef, para_check)
        except:
            return self.exit_codes.ERROR_CALC_PARAMETERS_INVALID  # pylint: disable=no-member

        para_check.set_multiple_values(
            NPT1=0,
            NPT3=0,
            NPOL=0,
        )

        updatenode = Dict(para_check.get_dict())
        updatenode.label = label + 'KKRparam_BS'
        updatenode.description = 'KKR parameter node extracted from remote_folder' + descr + ' as well as wf_parameter input node.'

        paranode_BS = update_params_wf(self.ctx.input_params_KKR, updatenode)
        self.ctx.BS_kkrparams = paranode_BS

    def get_BS(self):
        """
        submit the KkrCalcultion with the qdos settings for a bandstructure calculation
        """
        label = 'KKR BS calc.'
        BS_dict = self.ctx.BS_params_dict
        key_list = list(BS_dict)
        description = 'User defined BandStructure parameters '
        for key in key_list:
            description += f'{key}= {BS_dict[key]} ,'

        code = self.inputs.kkr
        remote = self.inputs.remote_data
        params = self.ctx.BS_kkrparams
        kpoints = self.ctx.BS_kpoints
        options = {
            'max_wallclock_seconds': self.ctx.max_wallclock_seconds,
            'resources': self.ctx.resources,
            'queue_name': self.ctx.queue,
        }
        if self.ctx.custom_scheduler_commands:
            options['custom_scheduler_commands'] = self.ctx.custom_scheduler_commands
        if self.ctx.append_text:
            options['append_text'] = self.ctx.append_text
        if self.ctx.prepend_text:
            options['prepend_text'] = self.ctx.prepend_text
        if self.ctx.additional_retrieve_list:
            options['additional_retrieve_list'] = self.ctx.additional_retrieve_list

        # get inputs for band structure calculation
        inputs = get_inputs_kkr(
            code, remote, options, label, description, parameters=params, serial=(not self.ctx.withmpi)
        )
        inputs.kpoints = kpoints

        # add nonco angles if found in the parent calculation or in the input
        if 'initial_noco_angles' in self.inputs:
            # overwrite nonco_angles from the input if given
            inputs['initial_noco_angles'] = self.inputs.initial_noco_angles
            self.report('used nonco angles from input to workflow')
        else:
            # extract from the parent calculation
            parent_calc = remote.get_incoming(node_class=KkrCalculation).first().node
            if 'initial_noco_angles' in parent_calc.inputs:
                noco_angles = extract_noco_angles(
                    fix_dir_threshold=Float(1e-6),  # make small enough
                    old_noco_angles=parent_calc.inputs.initial_noco_angles,
                    last_retrieved=parent_calc.outputs.retrieved
                )
                # set nonco angles (either from input or from output if it was updated)
                if noco_angles == {}:
                    noco_angles = parent_calc.inputs.initial_noco_angles
                self.report(f'extract nonco angles and use from parent ({noco_angles})')

        # LDA+U settings
        if 'settings_LDAU' in self.inputs:
            self.report('Add settings_LDAU input node')
            inputs.settings_LDAU = self.inputs.settings_LDAU

        BS_run = self.submit(KkrCalculation, **inputs)
        self.ctx.last_calc = BS_run

        return ToContext(BS_run=BS_run)

    def return_results(self):
        """
        Collect results, parse BS_calc output and link output nodes to workflow node
        """

        caching_info = f'INFO: cache_source of BS calc node: {self.ctx.BS_run.get_cache_source}'
        self.report(caching_info)

        if not self.ctx.BS_run.is_finished_ok:
            self.ctx.successful = False
            error = f'ERROR BS calculation failed somehow it is in state {self.ctx.BS_run.process_state}'
            self.report(error)
            self.ctx.errors.append(error)
            return self.exit_codes.ERROR_BS_CALC_FAILED  # pylint: disable=no-member

        # create dict to store results of workflow output
        outputnode_dict = {}
        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['workflow_version'] = self._wf_version
        outputnode_dict['withmpi'] = self.ctx.withmpi
        outputnode_dict['resources'] = self.ctx.resources
        outputnode_dict['max_wallclock_seconds'] = self.ctx.max_wallclock_seconds
        outputnode_dict['queue_name'] = self.ctx.queue
        outputnode_dict['custom_scheduler_commands'] = self.ctx.custom_scheduler_commands
        outputnode_dict['BS_params'] = self.ctx.BS_params_dict
        if 'kpoints' not in self.inputs:
            outputnode_dict['structure_type'] = self.ctx.structure_data
        outputnode_dict['BS_wf_description'] = self.ctx.description_wf
        outputnode_dict['BS_wf_label'] = self.ctx.label_wf
        try:
            outputnode_dict['nspin'] = self.ctx.BS_run.res.nspin
        except:
            error = 'ERROR: nspin not extracted'
            self.report(error)
            self.ctx.successful = False
            self.ctx.errors.append(error)
        outputnode_dict['successful'] = self.ctx.successful
        outputnode_dict['list_of_errors'] = self.ctx.errors

        # create output node with data-provenance
        outputnode = Dict(outputnode_dict)
        outputnode.label = 'kkr_BS_wc_results'
        outputnode.description = 'Contains the info of the WC'

        self.report('INFO: create Banstructure results nodes')
        try:
            self.report(
                f'INFO: create Bandstructure results nodes. BS calc retrieved node={self.ctx.BS_run.outputs.retrieved}'
            )
            has_BS_run = True
        except AttributeError as e:
            self.report('ERROR: No Bandstructure calc retrieved node found')
            self.report(f'Caught AttributeError {e}')
            return self.exit_codes.ERROR_BS_CALC_FAILED  # pylint: disable=no-member

        if has_BS_run:
            BS_retrieved = self.ctx.BS_run.outputs.retrieved

        ef = self.ctx.fermi_energy  # in Ry unit
        kpoints = self.ctx.BS_kpoints

        # Here outdict dictionary has been created to set the Dict result_wf, BS_data
        # to the output(spec.output) of the wf
        outdict = {}
        if has_BS_run:
            ArraData = parse_BS_data(BS_retrieved, Float(ef), kpoints)
            outdict['BS_Data'] = ArraData['BS_Data']

        # link to the BS output nodes
        link_nodes = outdict.copy()

        outdict['results_wf'] = create_out_dict_node(outputnode, **link_nodes)

        # create links to output nodes
        for link_name, node in outdict.items():
            self.out(link_name, node)

        self.report('INFO: done with BS_workflow!\n')


def set_energy_params(econt_new, ef, para_check):
    """
    set energy contour values to para_check
    internally convert from relative eV units to absolute Ry units
    """
    evscal = get_Ry2eV()

    for key, val in econt_new.items():
        if key in ['kmesh', 'BZDIVIDE', 'KMESH', 'bzdivide']:
            key = 'BZDIVIDE'
        elif key in ['nepts', 'NPT2']:
            key = 'NPT2'
            # also add IEMXD which has to be big enough
            para_check.set_value('IEMXD', val, silent=True)
        elif key in ['emin', 'EMIN']:
            key = 'EMIN'
            val = (ef + val / evscal)  # converting the Energy value to Ry while the fermi_energy in Ry
        elif key in ['emax', 'EMAX']:
            key = 'EMAX'
            val = (ef + val / evscal)  # Converting to the Ry (unit of the energy)
        elif key in ['tempr', 'TEMPR']:
            key = 'TEMPR'
        elif key in ['RCLUSTZ', 'rclustz']:
            key = 'RCLUSTZ'
        para_check.set_value(key, val, silent=True)

    # set the rest of the DOS contour
    para_check.set_multiple_values(
        NPT1=0,
        NPT3=0,
        NPOL=0,
        use_semi_circle_contour=False,  # this is needed to get a DOS contour
    )

    # set KPOIBZ to match BZDIVIDE setting
    # this is only done if 'KPOIBZ' is not given already in the input
    bzdiv = para_check.get_value('BZDIVIDE')
    if bzdiv is not None and 'KPOIBZ' not in econt_new:
        para_check.set_value('KPOIBZ', np.prod(bzdiv), silent=True)
    # we need to make sure to deactivate the semi-circle contour, otherwise DOS contour is not used
    para_check.set_value('<USE_SEMI_CIRCLE_CONTOUR>', False, silent=True)

    return para_check


@calcfunction
def parse_BS_data(retrieved_folder, fermi_level, kpoints):
    """
    parse the qdos files from the retreived folderand save as ArrayData
    """
    # conversion factor from Ry to eV
    eVscale = get_Ry2eV()

    retrieved_list = retrieved_folder.list_object_names()
    qdos_file_list = [i for i in retrieved_list if 'qdos.' in i]
    q_vec_file = 'qvec.dat'

    with retrieved_folder.open(q_vec_file) as file_opened:
        q_vec = np.loadtxt(file_opened, skiprows=1)

    for icount, fname in enumerate(qdos_file_list):
        with retrieved_folder.open(fname) as _f:
            loaded_file = np.loadtxt(_f)
            if icount == 0:
                total_qdos = loaded_file
            else:
                total_qdos[:, 5:] += loaded_file[:, 5:]

    ef = fermi_level.value  # in Ry unit
    total_qdos[:, 0] = (total_qdos[:, 0] - ef) * eVscale
    eng_points = set(total_qdos[:, 0])
    eng_points = np.sort(list(eng_points))
    no_eng_points = len(eng_points)

    qdos_intensity = np.ndarray(shape=(no_eng_points, len(q_vec)))
    for ne in range(np.shape(qdos_intensity)[0]):
        nk = np.shape(qdos_intensity)[1]
        # sum up all l-channels (5 is only the s-channel!)
        qdos_intensity[ne, :] = np.sum(total_qdos[ne * nk:(ne + 1) * nk, 5:], axis=1) / eVscale

    qdos_intensity = qdos_intensity.T  # setting eng-kpts corresponds to x-y asix
    q_vec = np.asarray(q_vec)  # converting q_vec into array
    eng_points = (np.asarray(eng_points))  # converting eng_popints into array in Ry unit

    # To save into the ArrayData
    array = ArrayData()
    array.set_array('BlochSpectralFunction', qdos_intensity)
    array.set_array('Kpts', q_vec)
    array.set_array('energy_points', eng_points)
    if kpoints.labels is not None:
        klbl_dict = dict(kpoints.labels)  # Special k-points
        array.extras['k-labels'] = klbl_dict

    return {'BS_Data': array}



# Workflow for impurity BdG calculation from converged normal state impurity portential and BdG host calculation

from aiida.engine import WorkChain, ToContext, if_
from aiida.orm import Dict, RemoteData, Code, CalcJobNode, WorkChainNode, Float, Bool, XyData, SinglefileData
from aiida_kkr.workflows import kkr_imp_wc, kkr_imp_dos_wc
from aiida_kkr.tools.find_parent import get_calc_from_remote
from aiida_kkr.tools.common_workfunctions import test_and_get_codenode

__copyright__ = (u'Copyright (c), 2022, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.1.3'
__contributors__ = (u'David Antognini Silva, Philipp Rüßmann')

# TODO: add _wf_default parameters and activate get_wf_defaults method
# TODO: add input interdependencies in the workchain description


class kkrimp_BdG_wc(WorkChain):
    """
    Workchain for one-shot BdG impurity DOS calculation from a converged normal state host calculation.
    1) Normal state impurity potential scf
    2) One-shot BdG DOS calcluation
        a) Writing out of the BdG kkrflex DOS files
        b) One-shot BdG impurity DOS
    The intermediate steps (1 & 2a) can be skipped by providing the corresponding nodes as inputs to the workflow.

    inputs::
        :param options: (Dict), computer options
        :param impurity_info: (Dict), information about the impurity cluster
        :param voronoi: (Code), Voronoi code for creating the impurity starting potential
        :param kkr: (Code), KKR host code for the writing out kkrflex files
        :param kkrimp: (Code), KKR impurity code for the normal state impurity scf and BdG impurity DOS calculation
        :param BdG_settings: (Dict), set BdG parameters
        :param imp_scf.startpot: (SinglefileData), pre-converged impurity potential used to start the impurity scf calculation
        :param imp_scf.converged_potential: (SinglefileData), converged impurity potential, skips the impurity scf calculation if provided
        :param imp_scf.wf_parameters: (Dict), parameters for the kkr impurity scf
        :param imp_scf.gf_writeout.params_kkr_overwrite: (Dict), set some input parameters of the KKR calculation for the GF writeout step of impurity scf workchain
        :param imp_scf.gf_writeout.options: (Dict), computer settings
        :param imp_scf.scf.params_overwrite: (Dict), set some input parameters of the KKR impurity scf
        :param imp_scf.options: (Dict), computer settings
        :param imp_scf.remote_data_host: (RemoteData), parent folder of converged host normal state KkrCalculation

        :param dos.wf_parameters: (Dict), parameters for the DOS calculation
        :param dos.gf_dos_remote: (RemoteData), node of precomputed host GF for DOS energy contour
        :param dos.gf_writeout.params_kkr_overwrite: (Dict), set some input parameters of the KKR calculation for the GF writeout step of imßpurity dos workchain
        :param dos.gf_writeout.host_remote: (RemoteData), parent folder of kkrflex writeout step for DOS calculation
        :param dos.gf_writeout.kkr: (Code), KKR code for writing out of kkrflex files for impurity DOS calculation
        :param dos.gf_writeout.options: (Dict), computer settings
        :param dos.options: (Dict), computer settings

    returns::
        :return workflow_info: (Dict), Information on workflow results
        :return output_parameters: (Dict), output parameters of the workflow
        :return dos_data: (XyData), impurity DOS data output node
        :return dos_data_interpol: (XyData), interpolated DOS data output node
        :return impurity_potential: (SinglefileData), converged normal state impurity potential node
        :return gf_host_BdG: (RemoteData), kkrflex writeout step files of DOS calculation
    """

    _wf_version = __version__

    # _wf_default = {
    #     'default': None,  #Put in here default input parameters
    # }
    _options_default = {
        'max_wallclock_seconds': 36000,
        'resources': {
            'num_machines': 1
        },
        'withmpi': True,
        'queue_name': ''
    }

    # @classmethod
    # def get_wf_defaults(self, silent=False):
    #     """
    #     Return the default values of the workflow parameters (wf_parameters input node)
    #     """
    #     if not silent:
    #         print(f'Version of the kkrimp_BdG_wc workflow: {self._wf_version}')
    #     return self._wf_default.copy()

    @classmethod
    def define(cls, spec):
        """
        Layout of the workflow, defines the input nodes and the outline of the workchain
        """
        super(kkrimp_BdG_wc, cls).define(spec)

        spec.input(
            'options',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(dict=cls._options_default),
            help=
            'Computer options (walltime etc.) passed onto KkrCalculation, fall back to settings from parent calculation if not given'
        )

        spec.input(
            'impurity_info',
            valid_type=Dict,
            required=False,
            help='Information of the impurity like position in the unit cell, screening cluster, atom type.'
        )

        spec.input('kkr', valid_type=Code, required=False, help='KKRhost code, needed to run the KkrCalculation')

        spec.input(
            'kkrimp', valid_type=Code, required=True, help='KKRimp code used to converge the impurity calculation'
        )

        spec.input(
            'voronoi',
            valid_type=Code,
            required=False,
            help='Voronoi code used to create the impurity starting potential.'
        )

        spec.input(
            'calc_DOS',
            valid_type=Bool,
            required=False,
            default=lambda: Bool(False),
            help='Set this to TRUE to calculate DOS'
        )

        spec.input(
            'BdG_settings',
            valid_type=Dict,
            default=lambda: Dict(
                dict={
                    # 'DELTA_BDG': 1e-4, # 1.36 meV # could be set in the future?
                    'USE_BdG': True,
                    'USE_E_SYMM_BDG': True,
                    # 'FIX_NONCO_ANGLES': True, # could be set in the future?
                }
            ),
            help='Define BdG parameters'
        )

        # expose inputs for impurity normal state scf
        spec.expose_inputs(
            kkr_imp_wc,
            namespace='imp_scf',
            include=(
                'startpot', 'wf_parameters', 'gf_writeout', 'scf.params_overwrite', 'scf.initial_noco_angles',
                'scf.settings_LDAU'
            )
        )
        spec.inputs['imp_scf']['gf_writeout']['kkr'].required = False
        spec.input('imp_scf.options', required=False, help='computer options for impurity scf step')
        spec.input(
            'imp_scf.converged_potential',
            valid_type=SinglefileData,
            required=False,
            help='Converged impurity potential. Skips impurity scf step.'
        )

        spec.input(
            'imp_scf.remote_data_host',
            valid_type=RemoteData,
            required=False,
            help='Parent folder of previously converged host normal state KkrCalculation'
        )

        spec.input(
            'imp_scf.remote_data_gf',
            valid_type=RemoteData,
            required=False,
            help='RemoteData node of precomputed host Green function'
        )

        # inputs for impurity BdG scf
        spec.expose_inputs(
            kkr_imp_wc,
            namespace='BdG_scf',
            include=('startpot', 'remote_data_gf', 'gf_writeout', 'scf.initial_noco_angles')
        )
        spec.inputs['BdG_scf']['gf_writeout']['kkr'].required = False
        spec.input('BdG_scf.options', required=False, help='computer options for BdG impurity scf step')

        spec.input(
            'BdG_scf.remote_data_host', required=False, help='Parent folder of previously converged BdG KkrCalculation'
        )

        # inputs for impurity dos
        spec.expose_inputs(
            kkr_imp_dos_wc,
            namespace='dos',
            include=('wf_parameters', 'gf_dos_remote', 'gf_writeout', 'initial_noco_angles')
        )

        spec.input(
            'dos.gf_writeout.host_remote',
            valid_type=RemoteData,
            required=False,
            help='Parent calculation from where the GF writeout starts. '
        )

        spec.input(
            'dos.gf_writeout.kkr',
            valid_type=Code,
            required=False,
            help='KKRhost code used to create DOS kkrflex files'
        )

        spec.input('dos.options', valid_type=Dict, required=False, help='Computer options for DOS step')

        # Here outputs are defined

        spec.output('workflow_info', valid_type=Dict, required=False)
        spec.output('output_parameters', valid_type=Dict, required=False)
        spec.output('dos_data', required=False, valid_type=XyData)
        spec.output('dos_data_interpol', required=False, valid_type=XyData)
        spec.output('dos_data_lm', required=False, valid_type=XyData)
        spec.output('dos_data_lm_interpol', required=False, valid_type=XyData)
        spec.output('impurity_potential', valid_type=SinglefileData)
        spec.output('gf_host_BdG', valid_type=RemoteData, required=False)

        # Here outlines are being specified
        spec.outline(
            # For initializing workflow
            cls.start,
            cls.validate_input,
            if_(cls.do_imp_pot_calc)(cls.imp_pot_calc),
            if_(cls.do_BdG_scf)(cls.imp_BdG_calc),
            if_(cls.do_calc_DOS)(cls.DOS_calc),
            cls.results
        )

        # Define all possible error messages
        spec.exit_code(
            100, 'ERROR_KKRCODE_NOT_CORRECT', 'The code you provided for kkr does not use the plugin kkr.kkr'
        )
        spec.exit_code(
            101, 'ERROR_KKRIMPCODE_NOT_CORRECT', 'The code you provided for kkrimp does not use the plugin kkr.kkrimp'
        )
        spec.exit_code(
            102, 'ERROR_VORONOICODE_NOT_CORRECT',
            'The code you provided for voronoi does not use the plugin kkr.voronoi'
        )

        spec.exit_code(200, 'ERROR_INVALID_PARENT', 'Parent calculation is not valid')

        #Now here we define all the functions from the outline

    def start(self):
        """
        Set up context of the workflow
        """
        self.report(f'INFO: started KKR BdG impurity version {self._wf_version}')

    def validate_input(self):
        """
        validate inputs
        """

        # validate for kkr code
        if 'kkr' in self.inputs:
            try:
                test_and_get_codenode(self.inputs.kkr, 'kkr.kkr', use_exceptions=True)
            except ValueError:
                return self.exit_codes.ERROR_KKRCODE_NOT_CORRECT  # pylint: disable=no-member

        # validate for kkrimp code, always done because this is a required input
        try:
            test_and_get_codenode(self.inputs.kkrimp, 'kkr.kkrimp', use_exceptions=True)
        except ValueError:
            return self.exit_codes.ERROR_KKRIMPCODE_NOT_CORRECT  # pylint: disable=no-member

        # validate for voronoi code
        if 'voronoi' in self.inputs:
            try:
                test_and_get_codenode(self.inputs.voronoi, 'kkr.voro', use_exceptions=True)
            except ValueError:
                return self.exit_codes.ERROR_VORONOICODE_NOT_CORRECT  # pylint: disable=no-member

    def do_imp_pot_calc(self):
        """
        run impurity potential calculation only if converged impurity potential not provided already
        """
        if not 'converged_potential' in self.inputs.imp_scf:
            return True

    def imp_pot_calc(self):
        """
        run normal state impurity scf calculation
        """

        builder = kkr_imp_wc.get_builder()

        builder.impurity_info = self.inputs.impurity_info
        builder.voronoi = self.inputs.voronoi
        builder.kkr = self.inputs.kkr
        builder.kkrimp = self.inputs.kkrimp
        if 'remote_data_gf' in self.inputs.imp_scf:
            builder.remote_data_gf = self.inputs.imp_scf.remote_data_gf
        else:
            builder.remote_data_host = self.inputs.imp_scf.remote_data_host
        builder.wf_parameters = self.inputs.imp_scf.wf_parameters
        if 'options' in self.inputs.imp_scf:
            builder.options = self.inputs.imp_scf.options
        else:
            builder.options = self.inputs.options
        if 'initial_noco_angles' in self.inputs.imp_scf:
            builder.scf.initial_noco_angles = self.inputs.imp_scf.initial_noco_angles  # pylint: disable=no-member

        if 'gf_writeout' in self.inputs.imp_scf:
            if 'options' in self.inputs.imp_scf.gf_writeout:
                builder.gf_writeout.options = self.inputs.imp_scf.gf_writeout.options  # pylint: disable=no-member
            if 'kkr' in self.inputs.imp_scf.gf_writeout:
                builder.kkr = self.inputs.imp_scf.gf_writeout.kkr
            if 'params_kkr_overwrite' in self.inputs.imp_scf.gf_writeout:
                builder.params_kkr_overwrite = self.inputs.imp_scf.gf_writeout.params_kkr_overwrite
        builder.gf_writeout.kkr = builder.kkr  # pylint: disable=no-member

        if 'params_overwrite' in self.inputs.imp_scf.scf:
            builder.scf.params_overwrite = self.inputs.imp_scf.scf.params_overwrite  # pylint: disable=no-member

        if 'startpot' in self.inputs.imp_scf:
            builder.startpot = self.inputs.imp_scf.startpot

        if 'settings_LDAU' in self.inputs.imp_scf.scf:
            builder.scf.settings_LDAU = self.inputs.imp_scf.scf.settings_LDAU  # pylint: disable=no-member

        imp_calc = self.submit(builder)

        return ToContext(last_imp_calc=imp_calc)

    def imp_BdG_calc(self):
        """
        run BdG one-shot impurity calculation
        """

        builder = kkr_imp_wc.get_builder()
        if 'impurity_info' in self.inputs:
            builder.impurity_info = self.inputs.impurity_info
        builder.voronoi = self.inputs.voronoi
        builder.kkr = self.inputs.kkr
        builder.kkrimp = self.inputs.kkrimp

        if 'startpot' in self.inputs.imp_scf:
            builder.startpot = self.inputs.imp_scf.startpot
        else:
            builder.startpot = self.ctx.last_imp_calc.outputs.converged_potential

        if 'remote_data_gf' in self.inputs.BdG_scf:
            builder.remote_data_gf = self.inputs.BdG_scf.remote_data_gf

        if 'gf_writeout' in self.inputs.BdG_scf:
            if 'kkr' in self.inputs.BdG_scf.gf_writeout:
                builder.kkr = self.inputs.BdG_scf.gf_writeout.kkr
            if 'params_kkr_overwrite' in self.inputs.BdG_scf.gf_writeout:
                builder.params_kkr_overwrite = self.inputs.BdG_scf.gf_writeout.params_kkr_overwrite
        if 'kkr' in self.inputs:
            builder.gf_writeout.kkr = builder.kkr  # pylint: disable=no-member
        if 'initial_noco_angles' in self.inputs.BdG_scf:
            builder.scf.initial_noco_angles = self.inputs.BdG_scf.initial_noco_angles  # pylint: disable=no-member

        builder.remote_data_host = self.inputs.BdG_scf.remote_data_host

        if 'options' in self.inputs.BdG_scf:
            builder.options = self.inputs.BdG_scf.options
        else:
            if 'options' in self.inputs.imp_scf:
                builder.options = self.inputs.imp_scf
            else:
                builder.options = self.inputs.options

        settings = kkr_imp_wc.get_wf_defaults()[1]
        settings['strmix'] = 0.01
        settings['aggrmix'] = 0.01
        settings['nsteps'] = 1
        settings['kkr_runmax'] = 1
        settings['mag_init'] = True
        builder.wf_parameters = Dict(dict=settings)
        BdG_params = self.inputs.BdG_settings.get_dict()
        builder.scf.params_overwrite = Dict(dict=BdG_params)  # pylint: disable=no-member

        imp_calc_BdG = self.submit(builder)

        return ToContext(last_imp_calc_BdG=imp_calc_BdG)

    def do_BdG_scf(self):
        """
        run BdG scf step only if BdG impurity potential not provided and DOS calculation is not planned
        """
        # BdG scf not implemented in the KKRimp code yet
        return False
        # if (not 'startpot' in self.inputs.BdG_scf) and (not self.inputs.calc_DOS):
        # return True

    def do_calc_DOS(self):
        """
        check if DOS calculation should be done
        """
        if self.inputs.calc_DOS:
            return True

    def DOS_calc(self):
        """
        run DOS calculation
        """
        from aiida_kkr.workflows import kkr_imp_dos_wc

        builder = kkr_imp_dos_wc.get_builder()

        if 'kkr' in self.inputs:
            builder.kkr = self.inputs.kkr
        builder.kkrimp = self.inputs.kkrimp

        #define computer options
        if 'options' in self.inputs.dos:
            builder.options = self.inputs.dos.options
        else:
            if 'options' in self.inputs.BdG_scf:
                builder.options = self.inputs.BdG_scf.options
            else:
                if 'options' in self.inputs.imp_scf:
                    builder.options = self.inputs.imp_scf.options
                else:
                    builder.options = self.inputs.options

        # set nonco angles
        if 'initial_noco_angles' in self.inputs.dos:
            builder.initial_noco_angles = self.inputs.dos.initial_noco_angles

        # skip BdG step and just use the starting potential instead?
        # faster and same accuracy?!
        if 'startpot' in self.inputs.BdG_scf:
            builder.imp_pot_sfd = self.inputs.BdG_scf.startpot
        else:
            if not self.inputs.calc_DOS:
                builder.imp_pot_sfd = self.ctx.last_imp_calc_BdG.outputs.converged_potential
            else:
                if 'converged_potential' in self.inputs.imp_scf:
                    builder.imp_pot_sfd = self.inputs.imp_scf.converged_potential
                else:
                    builder.imp_pot_sfd = self.ctx.last_imp_calc.outputs.converged_potential

        if 'impurity_info' in self.inputs:
            builder.impurity_info = self.inputs.impurity_info

        builder.wf_parameters = Dict(dict=kkr_imp_dos_wc.get_wf_defaults())

        if 'wf_parameters' in self.inputs.dos:
            builder.wf_parameters = self.inputs.dos.wf_parameters

        BdG_params = self.inputs.BdG_settings.get_dict()
        builder.BdG.params_overwrite = Dict(dict=BdG_params)  # pylint: disable=no-member

        if 'gf_writeout' in self.inputs.dos:
            # set optional inputs
            if 'kkr' in self.inputs.dos.gf_writeout:
                builder.kkr = self.inputs.dos.gf_writeout.kkr
            if 'params_kkr_overwrite' in self.inputs.dos.gf_writeout:
                builder.gf_writeout.params_kkr_overwrite = self.inputs.dos.gf_writeout.params_kkr_overwrite  # pylint: disable=no-member
            if 'host_remote' in self.inputs.dos.gf_writeout:
                builder.host_remote = self.inputs.dos.gf_writeout.host_remote
            if 'options' in self.inputs.dos.gf_writeout:
                builder.gf_writeout.options = self.inputs.dos.gf_writeout.options  # pylint: disable=no-member
        if 'kkr' in self.inputs:
            builder.gf_writeout.kkr = builder.kkr  # pylint: disable=no-member
        if 'gf_dos_remote' in self.inputs.dos:
            builder.gf_dos_remote = self.inputs.dos.gf_dos_remote

        DOS_calc = self.submit(builder)

        return ToContext(DOS_node=DOS_calc)

    def results(self):
        """
        return the results nodes of the workchain
        """
        if self.inputs.calc_DOS:
            if 'dos_data' in self.ctx.DOS_node.outputs:
                self.out('dos_data', self.ctx.DOS_node.outputs.dos_data)
            if 'dos_data_interpol' in self.ctx.DOS_node.outputs:
                self.out('dos_data_interpol', self.ctx.DOS_node.outputs.dos_data_interpol)
            if 'dos_data_lm' in self.ctx.DOS_node.outputs:
                self.out('dos_data_lm', self.ctx.DOS_node.outputs.dos_data_lm)
            if 'dos_data_interpol_lm' in self.ctx.DOS_node.outputs:
                self.out('dos_data_interpol_lm', self.ctx.DOS_node.outputs.dos_data_interpol_lm)

        if self.do_BdG_scf():
            if 'startpot' not in self.inputs.BdG_scf:
                self.out('workflow_info', self.ctx.last_imp_calc_BdG.outputs.workflow_info)
                self.out('output_parameters', self.ctx.last_imp_calc_BdG.outputs.last_calc_output_parameters)
                if 'remote_data_gf' not in self.inputs.BdG_scf:
                    self.out('gf_host_BdG', self.ctx.last_imp_calc_BdG.outputs.remote_data_gf)
                else:
                    self.out('gf_host_BdG', self.inputs.BdG_scf.remote_data_gf)

        if 'converged_potential' not in self.inputs.imp_scf:
            self.out('impurity_potential', self.ctx.last_imp_calc.outputs.converged_potential)
        else:
            self.out('impurity_potential', self.inputs.imp_scf.converged_potential)
